;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "org-transclusion-blocks" "0.4.0.0.20251218.3"
  "Component-based transclusion framework."
  '((emacs            "29.1")
    (org-transclusion "1.4.0")
    (org              "9.7"))
  :url "https://github.com/gggion/org-transclusion-blocks"
  :commit "919f2351772f85271e7aa5112d06c066006974ce"
  :revdesc "919f2351772f"
  :keywords '("hypermedia" "vc")
  :maintainers '(("Gino Cornejo" . "gggion123@gmail.com")))
